/*
  master.c
  Kernel module for 2.2.x Kernels
  
  This file is part of drbd Philipp Reisner.

  drbd is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  drbd is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with drbd; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifdef HAVE_AUTOCONF
#include <linux/autoconf.h>
#endif
#ifdef CONFIG_MODVERSIONS
#include <linux/modversions.h>
#endif  

#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/init.h>
#include <linux/slab.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#include "drbd.h"

struct Drbd_Conf
{
  struct ioctl_drbd_config conf;
  int                      count;
  struct socket*           sock;
  kdev_t		   lo_device;
  struct file*             lo_file;
  Drbd_State               state;
  Drbd_CState              cstate;
  unsigned long*           transfer_log;
  int                      tl_size;
  int                      send_cnt;
  int                      recv_cnt;
  int                      tpid;
  struct semaphore         tsem;
};

int drbd_send(struct socket* sock,Drbd_Packet_Cmd cmd,int len,
	       unsigned long blk_nr,void * data);
int drbd_send_param(struct socket* sock,int minor);
void drbd_try_connect(int minor);
void start_drbdd(int minor);

static int drbd_proc_get_info(char * , char ** , off_t ,int , int );
static void drbd_dio_end(struct buffer_head *bh, int uptodate);

int drbd_init( void );
static int drbd_open(struct inode *inode, struct file *file);
static int drbd_close(struct inode *inode, struct file *file);
static int drbd_ioctl(struct inode *inode, struct file *file, 
		      unsigned int cmd, unsigned long arg);
static int drbd_fsync(struct file * file, struct dentry * dentry);
static void drbd_do_request( void ); 


/* #define MAJOR_NR 240 */
/* DUPLICATED */
#define MAJOR_NR 43
/* /DUPLICATED */
/* Using the major_nr of the network block device
   prevents us from deadlocking with no request entries
   left on all_requests...
   look out for NBD_MAJOR in ll_rw_blk.c */

#define DEVICE_ON(device)
#define DEVICE_OFF(device)
#define DEVICE_NR(device) (MINOR(device))
#include <linux/blk.h>

/* DUPLICATED */
#ifdef DEVICE_NAME
#undef DEVICE_NAME
#endif
#define DEVICE_NAME "drbd"
/* /DUPLICATED */

#ifdef DEVICE_REQUEST
#undef DEVICE_REQUEST
#endif
#define DEVICE_REQUEST drbd_do_request


#define MODULE_NAME DEVICE_NAME": "
#define MINOR_COUNT 2
#define DRBD_BLOCKSIZE 1024

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#if DBG
EXPORT_SYMBOL(drbd_ioctl);
EXPORT_SYMBOL(drbd_open);
EXPORT_SYMBOL(drbd_close);
EXPORT_SYMBOL(drbd_fsync);
EXPORT_SYMBOL(drbd_do_request);
EXPORT_SYMBOL(drbd_dio_end);
#endif
MODULE_AUTHOR("Philipp Reisner <e9525415@stud2.tuwien.ac.at>");
MODULE_DESCRIPTION("drbd - Network block device");

static int drbd_blocksizes[MINOR_COUNT];
static int drbd_sizes[MINOR_COUNT];
static struct Drbd_Conf drbd_conf[MINOR_COUNT];

static struct file_operations drbd_fops =
{
	NULL,                   /* lseek - default */
	block_read,             /* read - general block-dev read */
	block_write,            /* write - general block-dev write */
	NULL,                   /* readdir - bad */
	NULL,                   /* poll */
	drbd_ioctl,             /* ioctl */
	NULL,                   /* mmap */
	drbd_open,              /* open */
	NULL,			/* flush */
	drbd_close,             /* release */
	drbd_fsync,             /* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL,			/* revalidate */
	NULL                    /* lock */
};

/************************* PROC FS struff begin */
#include <linux/proc_fs.h>
  
struct proc_dir_entry drbd_proc_dir = 
{
  0, 4, "drbd",
  S_IFREG | S_IRUGO , 1, 0, 0,
  0, NULL,
  &drbd_proc_get_info, NULL,
  NULL,
  NULL, NULL
};

static int drbd_proc_get_info(char * buf , char ** start, off_t offset, 
			      int len, int unused)
{
  int rlen,i;
  static const char* cstate_names[6] =
  {
    "Unconfigured",
    "Unconnected", 
    "WFConnection",
    "WFReportParams",
    "Syncing", 
    "Connected" 
  };
  static const char* state_names[2] =
  {
    "Primary", 
    "Secondary"
  };


  rlen = sprintf(buf,
		 "version       : %d\n\n",
		 MOD_VERSION);    

  for (i = 0; i < MINOR_COUNT; i++) 
    {
      rlen = rlen + sprintf(buf+rlen,"%d: cs:%s st:%s s:%d r:%d\n",
			    i,
			    cstate_names[drbd_conf[i].cstate],
			    state_names[drbd_conf[i].state],
			    drbd_conf[i].send_cnt,
			    drbd_conf[i].recv_cnt);
    }


  return rlen;
}

/* PROC FS struff end */

/************************* The transfer log start */
/* TODO: Use more than one log.... */

#define TL_ORDER 6
#define TL_SIZE (1<<TL_ORDER)
#define TL_MASK (TL_SIZE-1)
#define NR_OF_LOGS 2

unsigned long transfer_log[TL_SIZE];


int tl_test_and_add(struct Drbd_Conf* mdev,unsigned long new_nr)
{
  int i;
  if(mdev->tl_size >= TL_SIZE) return 2;

  for(i=0;i<mdev->tl_size;i++)
    if(mdev->transfer_log[i] == new_nr) return 1;

  mdev->transfer_log[mdev->tl_size++] = new_nr;

  return 0;
}     

void tl_clear(struct Drbd_Conf* mdev)
{
  mdev->tl_size=0;
  
}
/* Transfer log end */ 

int drbd_send(struct socket* sock,Drbd_Packet_Cmd cmd,int len,
	      unsigned long blk_nr,void * data)
{
  mm_segment_t oldfs;
  struct msghdr msg;
  struct iovec iov[2];
  static Drbd_Packet packet_header;
  int err;

  if(! sock) return -1000;

  packet_header.magic = cpu_to_be32(DRBD_MAGIC);
  packet_header.command = cpu_to_be16(cmd);
  packet_header.length = cpu_to_be16(len);
  packet_header.block_nr = cpu_to_be64(blk_nr);

  iov[0].iov_base=(void*)&packet_header;
  iov[0].iov_len=sizeof(packet_header);
  iov[1].iov_base=data;
  iov[1].iov_len=len;

  msg.msg_iov=iov;
  msg.msg_iovlen=2;
  msg.msg_control=NULL;
  msg.msg_controllen=0;
  msg.msg_name=0;
  msg.msg_namelen=0;
  msg.msg_flags = 0; 
      
  lock_kernel();
  oldfs = get_fs(); set_fs(KERNEL_DS);
  err = sock_sendmsg(sock, &msg,
		     iov[1].iov_len+sizeof(packet_header));
  set_fs(oldfs);
  unlock_kernel();

  if(err != iov[1].iov_len+sizeof(packet_header) )
    printk( KERN_ERR DEVICE_NAME ": sock_sendmsg returned %d\n"
	    ,err);
  return err;
}

int drbd_send_param(struct socket* sock,int minor)
{
  Drbd_ParameterBlock param;
  struct inode *inode;  
  int err;

  inode = drbd_conf[minor].lo_file->f_dentry->d_inode;   
  if (blk_size[MAJOR(inode->i_rdev)])
    {
      param.my_size = 
	blk_size[MAJOR(inode->i_rdev)][MINOR(inode->i_rdev)];
    }
  else 
    printk( KERN_ERR DEVICE_NAME "LL Device has not size ?!?\n\n");

  param.my_state = drbd_conf[minor].state;
  
  err=drbd_send(sock,ReportParams,sizeof(Drbd_ParameterBlock),0,&param);
  if(err>0) drbd_conf[minor].send_cnt++;
  
  return err;
}

void drbd_try_connect(int minor)
{
  int err;
  struct socket* sock;

  if(drbd_conf[minor].sock) 
    {
      printk( KERN_ERR DEVICE_NAME ": There is already a socket!! \n");
      return;
    }

  err = sock_create(AF_INET, SOCK_STREAM, 0, &sock);
  if(err) {
    printk( KERN_ERR DEVICE_NAME ": sock_creat(..)=%d\n", err);
  }

  lock_kernel();
  err = sock->ops->connect(sock, 
		    (struct sockaddr*)drbd_conf[minor].conf.other_addr,
		    drbd_conf[minor].conf.other_addr_len, 0);
  unlock_kernel();

  if(err) 
    {
      sock_release(sock);

      printk( KERN_ERR DEVICE_NAME": Unable to connect to server (%d)\n", err);

      err = sock_create(AF_INET, SOCK_STREAM, 0, &sock);
      if(err) {
	printk( KERN_ERR DEVICE_NAME ": sock_creat(..)=%d\n", err);
      }
      lock_kernel();
      err = sock->ops->bind(sock, 
			 (struct sockaddr*)drbd_conf[minor].conf.my_addr, 
	                 drbd_conf[minor].conf.my_addr_len); 
      unlock_kernel();
      if(err){
	printk( KERN_ERR DEVICE_NAME ": Unable to bind (%d)\n", err);
	sock_release(sock);
	return;
      }
      drbd_conf[minor].cstate = WFConnection;
    }
  else 
    {
      err=drbd_send_param(sock,minor);
      printk( KERN_ERR DEVICE_NAME": drbd_send_param()=%d \n", err);
      drbd_conf[minor].cstate = WFReportParams;
    }

  drbd_conf[minor].sock=sock;

  if(err >= 0)
    {
      printk( KERN_ERR DEVICE_NAME": staring thread\n");
      start_drbdd(minor);
    }
  else
    {
      printk( KERN_ERR DEVICE_NAME": not staring thread\n");
    }
}


static void drbd_dio_end(struct buffer_head *bh, int uptodate)
{
  /* end_request() */
  struct request *req = bh->b_dev_id;
  
  if(!end_that_request_first(req, uptodate, DEVICE_NAME))
    end_that_request_last(req);

  kfree(bh);
}

static void drbd_do_request( void )
{
  char *strs[2];
  int minor;

  strs[0]="READ";
  strs[1]="WRITE";

    while ( TRUE )
    {
      /* INIT_REQUEST; */

	if (!CURRENT) {
	  /* send getAck Packet and wait for answer... */
	  	break; 
	} 
	if (MAJOR(CURRENT->rq_dev) != MAJOR_NR) 
		panic(DEVICE_NAME ": request list destroyed"); 
	if (CURRENT->bh) { 
		if (!buffer_locked(CURRENT->bh)) 
			panic(DEVICE_NAME ": block not locked"); 
	}
	  
	minor = MINOR(CURRENT->rq_dev);

	/*
	printk( KERN_ERR DEVICE_NAME ": do_request(cmd=%s,sec=%ld,"
		"nr_sec=%ld,cnr_sec=%ld,buf=%p,min=%d)",
		strs[CURRENT->cmd == READ ? 0 : 1],CURRENT->sector,
		CURRENT->nr_sectors,
		CURRENT->current_nr_sectors,
		CURRENT->buffer,minor);
	*/

	/* Send it out to the network */
	if(CURRENT->cmd == WRITE && drbd_conf[minor].state == Primary )
	  {
	    if(drbd_send(drbd_conf[minor].sock,Data,
			 CURRENT->current_nr_sectors<<9,
			 CURRENT->sector,CURRENT->buffer)>0)
	      drbd_conf[minor].send_cnt++;

	  }

	/* Do disk - IO */
	{
	  struct buffer_head *bh;
	  bh = kmalloc(sizeof(struct buffer_head),GFP_KERNEL);
	  if(!bh)
	    {
	      printk( KERN_ERR DEVICE_NAME ": coul'd not kmalloc()\n");
	      return;
	    }

	  memcpy(bh,CURRENT->bh,sizeof(struct buffer_head));

	  bh->b_dev        = drbd_conf[minor].lo_device;
	  bh->b_state      = (1<<BH_Req) | (1<<BH_Dirty);
	  bh->b_list       = BUF_LOCKED;
	  bh->b_dev_id     = CURRENT;
	  bh->b_end_io     = drbd_dio_end;	  

	  ll_rw_block(CURRENT->cmd,1,&bh);
	}
	CURRENT = CURRENT->next;	
    }
}


int drbd_fsync(struct file * file, struct dentry * dentry)
{
  int minor;
  int retval=0;

  /* retval=block_fsync(file,dentry); */
  retval=fsync_dev(dentry->d_inode->i_rdev);
  minor=MINOR(file->f_dentry->d_inode->i_rdev);
  printk( KERN_ERR DEVICE_NAME ": sync minor= %d \n",minor);

  /* send a sync request ... */

  /* wait for ACK packet ... */

  return retval;
}

static int drbd_ioctl(struct inode *inode, struct file *file, 
		      unsigned int cmd, unsigned long arg)
{
  int err;
  int minor;
  struct file *filp;

  minor=MINOR(inode->i_rdev);

  switch(cmd)
    {
    RO_IOCTLS(inode->i_rdev,arg);

    case DRBD_IOCTL_GET_VERSION:
      if((err=put_user(MOD_VERSION, (int *)arg))) return err;
      break;

    case DRBD_IOCTL_SET_STATE:
      drbd_conf[minor].state=(Drbd_State)arg;
      if(blk_size[MAJOR_NR][minor])
	/*	set_device_ro(MKDEV(MAJOR_NR,minor),
		drbd_conf[minor].state != Primary);      */
      printk( KERN_ERR DEVICE_NAME ": set_state(%d)\n",drbd_conf[minor].state);
      break;
      
    case DRBD_IOCTL_SET_CONFIG:
      printk( KERN_ERR DEVICE_NAME ": set_config()\n");
      if((err=copy_from_user(&drbd_conf[minor].conf,(void*)arg,
			sizeof(struct ioctl_drbd_config))))
	return err;

      filp = fget(drbd_conf[minor].conf.lower_device);
      if(!filp) return -EINVAL;
      inode = filp->f_dentry->d_inode;
      if(!S_ISBLK(inode->i_mode)) return -EINVAL;
      if((err = blkdev_open(inode, filp)))
	{      
	  printk( KERN_ERR DEVICE_NAME ": blkdev_open( %d:%d ,) returned %d\n",
		  MAJOR(inode->i_rdev),MINOR(inode->i_rdev),err );
	  return err;	  
	}

      if(drbd_conf[minor].lo_file)
	{
	  blkdev_release(drbd_conf[minor].lo_file->f_dentry->d_inode);
	  fput(drbd_conf[minor].lo_file);
	}

      drbd_conf[minor].lo_device = inode->i_rdev;
      drbd_conf[minor].lo_file = filp;

      drbd_try_connect(minor);

      break;

    default:
      return -EINVAL;
    }
  return 0;
}


static int drbd_open(struct inode *inode, struct file *file)
{
  int minor;

  minor=MINOR(inode->i_rdev);

  printk( KERN_ERR DEVICE_NAME ": open(inode=%p,file=%p)"
	  "current=%p,minor=%d\n",inode,file,current,minor);

  if( drbd_conf[minor].count++ == 0)
    {
      printk( KERN_ERR DEVICE_NAME ": mallocing transfer log!\n");

      drbd_conf[minor].transfer_log = 
	kmalloc(sizeof(unsigned long) * TL_SIZE, GFP_KERNEL);
    }

  MOD_INC_USE_COUNT;

  return 0;
}

static void drbd_do_close(int minor)
{
  if( --drbd_conf[minor].count == 0)
    { /* this device is not opend by anyone, thus the thread is exiting ... */
      fsync_dev(MKDEV(MAJOR_NR,minor));
      /* must wait here? */
      if(drbd_conf[minor].sock)
	{
	  sock_release(drbd_conf[minor].sock);
	  drbd_conf[minor].sock=0;
	  printk( KERN_ERR DEVICE_NAME ": sock_release()minor=%d\n",minor);
	}
      if(drbd_conf[minor].lo_file)
	{
	  blkdev_release(drbd_conf[minor].lo_file->f_dentry->d_inode);
	  fput(drbd_conf[minor].lo_file);
	  drbd_conf[minor].lo_file=0;
	  drbd_conf[minor].lo_device=0;
	}
      blk_size[MAJOR_NR][minor] = 0;

      kfree(drbd_conf[minor].transfer_log);
      drbd_conf[minor].transfer_log=0;
    }
}


static int drbd_close(struct inode *inode, struct file *file)
{
  int minor;

  minor=MINOR(inode->i_rdev);

  printk( KERN_ERR DEVICE_NAME ": close(inode=%p,file=%p)"
	  "current=%p,minor=%d\n",inode,file,current,minor);

  drbd_do_close(minor);

  MOD_DEC_USE_COUNT;

  return 0;
}

                                                                             
__initfunc(int drbd_init( void ))
{
  int i;

  if(proc_register(&proc_root, &drbd_proc_dir))
    {
      printk(MODULE_NAME"unable to register proc file.\n");
      return -EIO;
    }

  if ( register_blkdev( MAJOR_NR, DEVICE_NAME, &drbd_fops ) )
    {
      printk( KERN_ERR DEVICE_NAME ": Unable to get major %d\n",
	      MAJOR_NR );
      return -EBUSY;
    }

  /* Initialize size arrays. */
  
  for (i = 0; i < MINOR_COUNT; i++) 
    {
      drbd_blocksizes[i] = DRBD_BLOCKSIZE;
      drbd_sizes[i]=0;
      set_device_ro(MKDEV(MAJOR_NR,i),FALSE/*TRUE*/);     
      drbd_conf[i].count = 0;
      drbd_conf[i].sock = 0;
      drbd_conf[i].lo_file = 0;	
      drbd_conf[i].transfer_log = 0;
      drbd_conf[i].state = Secondary;
      drbd_conf[i].send_cnt=0;
      drbd_conf[i].recv_cnt=0;
    }
   
  blk_dev[ MAJOR_NR ].request_fn = DEVICE_REQUEST;
  blksize_size[ MAJOR_NR ] = drbd_blocksizes;
  blk_size[ MAJOR_NR ] = drbd_sizes; /* Size in Kb */

  return 0;
}

__initfunc(int init_module())
{
  printk(MODULE_NAME"module initialised. Version: %d\n",MOD_VERSION);

  return drbd_init();

}

void cleanup_module()
{
  int i;

  for (i = 0; i < MINOR_COUNT; i++) 
    {  

      fsync_dev(MKDEV(MAJOR_NR,i));

      if(drbd_conf[i].tpid)
	{
	  int err;
	  printk( KERN_ERR DEVICE_NAME ": sending sig!\n");

	  drbd_conf[i].tsem = MUTEX_LOCKED; 

	  err = kill_proc_info(SIGTERM,NULL,drbd_conf[i].tpid);

	  if(err==0)
	    down( &drbd_conf[i].tsem ); /* wait until the thread
					   has closed the socket */
	  else
	    printk( KERN_ERR DEVICE_NAME ": could not send signal\n");

	  printk( KERN_ERR DEVICE_NAME ": (pseudo) waitpid returned \n");

	  current->state = TASK_INTERRUPTIBLE;
	  schedule_timeout(HZ/4);  

	  /*
	    This would be the *nice* solution, but it crashed
	    my machine...
	    
	  struct task_struct *p;
	  read_lock(&tasklist_lock);
	  p = find_task_by_pid(drbd_conf[i].tpid);
	  p->p_pptr = current;
	  errno = send_sig_info(SIGTERM, NULL, p);
	  read_unlock(&tasklist_lock);
	  interruptible_sleep_on(&current->wait_chldexit);
	  */

	}
    
      if(drbd_conf[i].sock)
	{
	  sock_release(drbd_conf[i].sock);
	  drbd_conf[i].sock=0;
	}
      
      if(drbd_conf[i].lo_file)
	{
	  blkdev_release(drbd_conf[i].lo_file->f_dentry->d_inode);
	  fput(drbd_conf[i].lo_file);
	  drbd_conf[i].lo_file=0;
	  drbd_conf[i].lo_device=0;
	}
      
    }

  if ( unregister_blkdev( MAJOR_NR, DEVICE_NAME ) != 0 )
    printk(MODULE_NAME"unregister of device failed\n");

  blk_dev[ MAJOR_NR ].request_fn = NULL;
  blksize_size[ MAJOR_NR ] = NULL;
  blk_size[ MAJOR_NR ] = NULL;
  
  proc_unregister(&proc_root, drbd_proc_dir.low_ino);

}


/************************* Receiving part */
#define min(a,b) ( (a) < (b) ? (a) : (b) )

/* struct semaphore test_sem; */

struct socket* drbd_accept(struct socket* sock)
{
  struct socket *newsock;
  int err=0;

  lock_kernel();

  err=sock->ops->listen(sock, 5);
  if(err) goto out;

  if (!(newsock = sock_alloc())) 
    goto out;

  newsock->type = sock->type;

  err = sock->ops->dup(newsock, sock);
  if (err < 0) 
    goto out_release;

  err = newsock->ops->accept(sock, newsock, 0);
  if (err < 0)
    goto out_release;

  unlock_kernel();
  return newsock;

 out_release:
  sock_release(newsock);
 out:
  unlock_kernel();
  printk( KERN_ERR DEVICE_NAME ": accept failed! %d\n",err);
  return 0;
}

int drbd_recv(struct socket *sock, void * ubuf, size_t size)
{
  mm_segment_t oldfs;

  struct iovec iov;
  struct msghdr msg;
  int err=0;

  msg.msg_control=NULL;
  msg.msg_controllen=0;
  msg.msg_iovlen=1;
  msg.msg_iov=&iov;
  iov.iov_len=size;
  iov.iov_base=ubuf;
  msg.msg_name=NULL;
  msg.msg_namelen=0;
  msg.msg_flags = 0;

  lock_kernel();
  oldfs = get_fs(); set_fs(KERNEL_DS);
  /* Check if this get_fs() / set_fs stuff is needed for a kernel_thread */

  err=sock_recvmsg(sock, &msg, size, MSG_WAITALL );

  set_fs(oldfs);
  unlock_kernel();
  if(err < 0 ) 
    printk( KERN_ERR DEVICE_NAME ": sock_recvmsg returned %d\n"
	    ,err);

  return err;
}

int drbdd(void *arg)
{
  struct socket* my_sock;
  Drbd_Packet packet_header;
  int minor = (int)((long)arg);
  kdev_t ll_dev = drbd_conf[minor].lo_file->f_dentry->d_inode->i_rdev;
  int rr;

  drbd_conf[minor].count++;  

  exit_mm(current);    /* give up UL-memory context */
  exit_files(current); /* give up open filedescriptors */
  current->session = 1;
  current->pgrp = 1;
  current->fs->umask = 0;

  sprintf(current->comm, "drbdd_%d",minor);

  down( &drbd_conf[minor].tsem ); /* wait until parent has written its
				     tpid variable */

  printk(KERN_ERR DEVICE_NAME ": thread living/m=%d\n",minor);

  switch(drbd_conf[minor].cstate)
    {
    case WFConnection:
      my_sock = drbd_accept(drbd_conf[minor].sock);
      if(!my_sock) goto out;

      sock_release(drbd_conf[minor].sock);
      drbd_conf[minor].sock=my_sock;

      if(drbd_send_param(my_sock,minor)<0)
	{
	  printk(KERN_ERR DEVICE_NAME ": could not send into accepted s.\n");
	  goto out;
	}
      break;
    case WFReportParams:
      my_sock = drbd_conf[minor].sock;
      break;
    default:
      printk(KERN_ERR DEVICE_NAME ": illegal starting state\n");
      goto out;
    }
  
  while(TRUE)
    {
      if( (rr=drbd_recv(my_sock, &packet_header, sizeof(Drbd_Packet))) <= 0) 
	break;
      drbd_conf[minor].recv_cnt++;

      if( be32_to_cpu(packet_header.magic) != DRBD_MAGIC )
	{
	  printk(KERN_ERR DEVICE_NAME ": no magic ??\n");
	  continue;
	}

      switch(be16_to_cpu(packet_header.command))
	{
	case Data: 
	  {
	    struct buffer_head * bh;
	    
	    /*
	    printk(KERN_ERR DEVICE_NAME ": recv Data "
		   "block_nr=%ld len=%d/m=%d\n",
		   be64_to_cpu(packet_header.block_nr),
		   (int)be16_to_cpu(packet_header.length),
		   minor); 
	    */
	    bh = getblk(MKDEV(MAJOR_NR,minor), 
			be64_to_cpu(packet_header.block_nr) >> 1, /* FIXME */
			DRBD_BLOCKSIZE);
	    if(!bh) 
	      {
		printk(KERN_ERR DEVICE_NAME ": getblk()=NULL /m=%d\n",minor); 
		break;
	      }

	    if( drbd_recv(my_sock,bh->b_data,
			  be16_to_cpu(packet_header.length)) <= 0 )
	      break;

	    mark_buffer_uptodate(bh, 1);
	    mark_buffer_dirty(bh, 0);
	    brelse(bh);	  
	    break;
	  }
	case Ack:  
	  printk(KERN_ERR DEVICE_NAME ": recv Ack/m=%d\n",minor); 
	  break;
	case SyncNow: 
	  printk(KERN_ERR DEVICE_NAME ": recv SyncNow/m=%d\n",minor); 
	  break;
	case ReportParams: 
	  {
	    Drbd_ParameterBlock param;

	    printk(KERN_ERR DEVICE_NAME ": recv ReportParams/m=%d\n",minor); 
	    if( drbd_recv(my_sock, &param, sizeof(Drbd_ParameterBlock)) <= 0) 
	      break;
	    
	    if (blk_size[MAJOR(ll_dev)])
	       {
		 blk_size[MAJOR_NR][minor] =
		   min(blk_size[MAJOR(ll_dev)][MINOR(ll_dev)],
		       param.my_size);
		 printk(KERN_ERR DEVICE_NAME ": reported size=%ld\n",
			param.my_size); 
		 /*		 set_device_ro(MKDEV(MAJOR_NR,minor),
				 drbd_conf[minor].state != Primary);      */
	       }
	    else
	      {
		blk_size[MAJOR_NR][minor] = 0;
		printk( KERN_ERR DEVICE_NAME "LL Device has not size ?!?\n\n");
	      }

	    drbd_conf[minor].cstate = Syncing;
	    break;
	  }
	default:
	  printk(KERN_ERR DEVICE_NAME ": unknown packet type!/m=%d\n",minor);
	  goto out;
	}
    }

 out:
  printk(KERN_ERR DEVICE_NAME ": thread dying/m=%d\n",minor);
  drbd_do_close(minor);
  drbd_conf[minor].tpid=0;
  up( &drbd_conf[minor].tsem );
  return 0;
}

void start_drbdd(int minor)
{
  int tpid;

  if(drbd_conf[minor].tpid==0)
    {
      drbd_conf[minor].tsem = MUTEX_LOCKED; 
      tpid = kernel_thread(drbdd, (void *)((long)minor), 0);

      if(tpid < 0) 
	{
	  printk(KERN_ERR DEVICE_NAME ": Couldn't start thread (%d)\n", tpid);
	  return;
	}

      printk(KERN_ERR DEVICE_NAME ": pid = %d\n",tpid);  
      drbd_conf[minor].tpid=tpid;
      up(&drbd_conf[minor].tsem);
    }
}

